import polib
from django.test import TestCase
from django.utils import timezone
from wagtail.core.models import Page, Locale

from wagtail_localize.models import (
    TranslationSource,
    String,
    StringTranslation,
    TranslationContext,
    Translation,
    UnknownString,
    UnknownContext,
    StringNotUsedInContext,
    CannotSaveDraftError,
)
from wagtail_localize.segments import RelatedObjectSegmentValue
from wagtail_localize.strings import StringValue
from wagtail_localize.test.models import TestPage, TestSnippet


def create_test_page(**kwargs):
    parent = kwargs.pop("parent", None) or Page.objects.get(id=1)
    page = parent.add_child(instance=TestPage(**kwargs))
    page_revision = page.save_revision()
    page_revision.publish()
    page.refresh_from_db()

    source, created = TranslationSource.get_or_create_from_instance(page)

    prepare_source(source)

    return page


def prepare_source(source):
    # Recurse into any related objects
    for segment in source.relatedobjectsegment_set.all():
        if not isinstance(segment, RelatedObjectSegmentValue):
            continue

        related_source, created = TranslationSource.get_or_create_from_instance(
            segment.get_instance(source.locale)
        )
        prepare_source(related_source)


class TestGetProgress(TestCase):
    def setUp(self):
        self.fr_locale = Locale.objects.create(language_code="fr")

        self.page = create_test_page(
            title="Test page",
            slug="test-slug",
            test_charfield="Test content",
            test_textfield="More test content"
        )
        self.source = TranslationSource.objects.get()
        self.translation = Translation.objects.create(
            source=self.source,
            target_locale=self.fr_locale,
        )

        self.test_charfield_context = TranslationContext.objects.get(path="test_charfield")
        self.test_textfield_context = TranslationContext.objects.get(path="test_textfield")

        self.test_content_string = String.objects.get(data="Test content")
        self.more_test_content_string = String.objects.get(data="More test content")

    def test_get_progress(self):
        progress = self.translation.get_progress()
        self.assertEqual(progress, (2, 0))

    def test_get_progress_with_translations(self):
        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=self.test_charfield_context,
            locale=self.fr_locale,
            data="Contenu de test",
        )

        StringTranslation.objects.create(
            translation_of=self.more_test_content_string,
            context=self.test_textfield_context,
            locale=self.fr_locale,
            data="Plus de contenu de test",
        )

        progress = self.translation.get_progress()
        self.assertEqual(progress, (2, 2))

    def test_get_progress_with_translations_in_different_locale(self):
        # Same as before but with a different locale. These shouldn't count
        de_locale = Locale.objects.create(language_code="de")

        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=self.test_charfield_context,
            locale=de_locale,
            data="Testinhalt",
        )

        StringTranslation.objects.create(
            translation_of=self.more_test_content_string,
            context=self.test_textfield_context,
            locale=de_locale,
            data="Mehr Testinhalte",
        )

        progress = self.translation.get_progress()
        self.assertEqual(progress, (2, 0))

    def test_get_progress_with_translations_in_different_context(self):
        # Same as before but swap the contexts of the translations. The translations shouldn't count
        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=self.test_textfield_context,
            locale=self.fr_locale,
            data="Contenu de test",
        )

        StringTranslation.objects.create(
            translation_of=self.more_test_content_string,
            context=self.test_charfield_context,
            locale=self.fr_locale,
            data="Plus de contenu de test",
        )

        progress = self.translation.get_progress()
        self.assertEqual(progress, (2, 0))


class TestExportPO(TestCase):
    def setUp(self):
        self.en_locale = Locale.objects.get(language_code="en")
        self.fr_locale = Locale.objects.create(language_code="fr")

        self.page = create_test_page(title="Test page", slug="test-page", test_charfield="This is some test content", test_textfield="This is some test content")
        self.source, created = TranslationSource.get_or_create_from_instance(self.page)

        self.translation = Translation.objects.create(
            source=self.source,
            target_locale=self.fr_locale,
        )

    def test_export_po(self):
        po = self.translation.export_po()

        self.assertEqual(po.metadata.keys(), {'POT-Creation-Date', 'MIME-Version', 'Content-Type', 'X-WagtailLocalize-TranslationID'})
        self.assertEqual(po.metadata['MIME-Version'], '1.0')
        self.assertEqual(po.metadata['Content-Type'], 'text/plain; charset=utf-8')
        self.assertEqual(po.metadata['X-WagtailLocalize-TranslationID'], str(self.translation.uuid))

        self.assertEqual(len(po), 2)
        self.assertEqual(po[0].msgid, "This is some test content")
        self.assertEqual(po[0].msgctxt, "test_charfield")
        self.assertEqual(po[0].msgstr, "")
        self.assertFalse(po[0].obsolete)

        # Test with the exact same content twice
        # This checks for https://github.com/wagtail/wagtail-localize/issues/253
        self.assertEqual(po[1].msgid, "This is some test content")
        self.assertEqual(po[1].msgctxt, "test_textfield")
        self.assertEqual(po[1].msgstr, "")
        self.assertFalse(po[1].obsolete)

    def test_export_po_with_translation(self):
        StringTranslation.objects.create(
            translation_of=String.objects.get(data="This is some test content"),
            context=TranslationContext.objects.get(path="test_charfield"),
            locale=self.fr_locale,
            data="Contenu de test",
        )

        po = self.translation.export_po()

        self.assertEqual(po.metadata.keys(), {'POT-Creation-Date', 'MIME-Version', 'Content-Type', 'X-WagtailLocalize-TranslationID'})
        self.assertEqual(po.metadata['MIME-Version'], '1.0')
        self.assertEqual(po.metadata['Content-Type'], 'text/plain; charset=utf-8')
        self.assertEqual(po.metadata['X-WagtailLocalize-TranslationID'], str(self.translation.uuid))

        self.assertEqual(len(po), 2)
        self.assertEqual(po[0].msgid, "This is some test content")
        self.assertEqual(po[0].msgctxt, "test_charfield")
        self.assertEqual(po[0].msgstr, "Contenu de test")
        self.assertFalse(po[0].obsolete)

        # Test with the exact same content twice
        # But note that this context doesn't have a translation!
        # This checks for https://github.com/wagtail/wagtail-localize/issues/253
        self.assertEqual(po[1].msgid, "This is some test content")
        self.assertEqual(po[1].msgctxt, "test_textfield")
        self.assertEqual(po[1].msgstr, "")
        self.assertFalse(po[1].obsolete)

    def test_export_po_with_obsolete_translation(self):
        obsolete_string = String.from_value(self.en_locale, StringValue("This is an obsolete string"))
        String.from_value(self.en_locale, StringValue("This is an obsolete string that was never translated"))

        StringTranslation.objects.create(
            translation_of=obsolete_string,
            context=TranslationContext.objects.get(path="test_charfield"),
            locale=self.fr_locale,
            data="Ceci est une chaîne obsolète",
        )

        po = self.translation.export_po()

        self.assertEqual(po.metadata.keys(), {'POT-Creation-Date', 'MIME-Version', 'Content-Type', 'X-WagtailLocalize-TranslationID'})
        self.assertEqual(po.metadata['MIME-Version'], '1.0')
        self.assertEqual(po.metadata['Content-Type'], 'text/plain; charset=utf-8')
        self.assertEqual(po.metadata['X-WagtailLocalize-TranslationID'], str(self.translation.uuid))

        # The non-obsolete strings come first
        self.assertEqual(len(po), 3)
        self.assertEqual(po[0].msgid, "This is some test content")
        self.assertEqual(po[0].msgctxt, "test_charfield")
        self.assertEqual(po[0].msgstr, "")
        self.assertFalse(po[0].obsolete)

        self.assertEqual(po[1].msgid, "This is some test content")
        self.assertEqual(po[1].msgctxt, "test_textfield")
        self.assertEqual(po[1].msgstr, "")
        self.assertFalse(po[1].obsolete)

        # Then the obsolete string
        self.assertEqual(po[2].msgid, "This is an obsolete string")
        self.assertEqual(po[2].msgctxt, "test_charfield")
        self.assertEqual(po[2].msgstr, "Ceci est une chaîne obsolète")
        self.assertTrue(po[2].obsolete)

        # Obsolete strings that never had a translation don't get exported


class TestImportPO(TestCase):
    def setUp(self):
        self.en_locale = Locale.objects.get(language_code="en")
        self.fr_locale = Locale.objects.create(language_code="fr")

        self.page = create_test_page(title="Test page", slug="test-page", test_charfield="This is some test content")
        self.source, created = TranslationSource.get_or_create_from_instance(self.page)

        self.translation = Translation.objects.create(
            source=self.source,
            target_locale=self.fr_locale,
        )

    def test_import_po(self):
        obsolete_string = String.from_value(self.en_locale, StringValue("This is an obsolete string"))
        StringTranslation.objects.create(
            translation_of=obsolete_string,
            context=TranslationContext.objects.get(path="test_charfield"),
            locale=self.fr_locale,
            data="Ceci est une chaîne obsolète",
        )

        po = polib.POFile(wrapwidth=200)
        po.metadata = {
            "POT-Creation-Date": str(timezone.now()),
            "MIME-Version": "1.0",
            "Content-Type": "text/plain; charset=utf-8",
            "X-WagtailLocalize-TranslationID": str(self.translation.uuid),
        }

        po.append(
            polib.POEntry(
                msgid="This is some test content",
                msgctxt="test_charfield",
                msgstr="Contenu de test",
            )
        )

        po.append(
            polib.POEntry(
                msgid="This is an obsolete string",
                msgctxt="test_charfield",
                msgstr="C'est encore une chaîne obsolète",
                obsolete=True
            )
        )

        warnings = self.translation.import_po(po)
        self.assertEqual(warnings, [])

        translation = StringTranslation.objects.get(translation_of__data="This is some test content")
        self.assertEqual(translation.context, TranslationContext.objects.get(path="test_charfield"))
        self.assertEqual(translation.locale, self.fr_locale)
        self.assertEqual(translation.data, "Contenu de test")

        # Obsolete strings still get updated
        translation = StringTranslation.objects.get(translation_of__data="This is an obsolete string")
        self.assertEqual(translation.context, TranslationContext.objects.get(path="test_charfield"))
        self.assertEqual(translation.locale, self.fr_locale)
        self.assertEqual(translation.data, "C'est encore une chaîne obsolète",)

    def test_import_po_deletes_translations(self):
        StringTranslation.objects.create(
            translation_of=String.objects.get(data="This is some test content"),
            context=TranslationContext.objects.get(path="test_charfield"),
            locale=self.fr_locale,
            data="Contenu de test",
        )

        StringTranslation.objects.create(
            translation_of=String.from_value(self.en_locale, StringValue("This is an obsolete string")),
            context=TranslationContext.objects.get(path="test_charfield"),
            locale=self.fr_locale,
            data="Ceci est une chaîne obsolète",
        )

        # Create an empty PO file
        po = polib.POFile(wrapwidth=200)
        po.metadata = {
            "POT-Creation-Date": str(timezone.now()),
            "MIME-Version": "1.0",
            "Content-Type": "text/plain; charset=utf-8",
            "X-WagtailLocalize-TranslationID": str(self.translation.uuid),
        }

        warnings = self.translation.import_po(po, delete=True)
        self.assertEqual(warnings, [])

        # Should delete both the translations
        self.assertFalse(StringTranslation.objects.exists())

    def test_import_po_with_invalid_translation_id(self):
        po = polib.POFile(wrapwidth=200)
        po.metadata = {
            "POT-Creation-Date": str(timezone.now()),
            "MIME-Version": "1.0",
            "Content-Type": "text/plain; charset=utf-8",
            "X-WagtailLocalize-TranslationID": 'foo',
        }

        po.append(
            polib.POEntry(
                msgid="This is some test content",
                msgctxt="test_charfield",
                msgstr="Contenu de test",
            )
        )

        warnings = self.translation.import_po(po)
        self.assertEqual(warnings, [])

        # Should delete both the translations
        self.assertFalse(StringTranslation.objects.exists())

    def test_warnings(self):
        String.from_value(self.en_locale, StringValue("This string exists in the database but isn't relevant to this object"))

        po = polib.POFile(wrapwidth=200)
        po.metadata = {
            "POT-Creation-Date": str(timezone.now()),
            "MIME-Version": "1.0",
            "Content-Type": "text/plain; charset=utf-8",
            "X-WagtailLocalize-TranslationID": str(self.translation.uuid),
        }

        po.append(
            polib.POEntry(
                msgid="This string exists in the database but isn't relevant to this object",
                msgctxt="test_charfield",
                msgstr="Contenu de test",
            )
        )

        po.append(
            polib.POEntry(
                msgid="This string doesn't exist",
                msgctxt="test_charfield",
                msgstr="Contenu de test",
            )
        )

        po.append(
            polib.POEntry(
                msgid="This is some test content",
                msgctxt="invalidcontext",
                msgstr="Contenu de test",
            )
        )

        warnings = self.translation.import_po(po)

        self.assertEqual(warnings, [
            StringNotUsedInContext(0, "This string exists in the database but isn't relevant to this object", "test_charfield"),
            UnknownString(1, "This string doesn't exist"),
            UnknownContext(2, "invalidcontext"),
        ])


class TestGetStatus(TestCase):
    def setUp(self):
        self.fr_locale = Locale.objects.create(language_code="fr")

        self.page = create_test_page(
            title="Test page",
            slug="test-slug",
            test_charfield="Test content",
            test_textfield="More test content"
        )
        self.source = TranslationSource.objects.get()
        self.translation = Translation.objects.create(
            source=self.source,
            target_locale=self.fr_locale,
        )

        self.test_charfield_context = TranslationContext.objects.get(path="test_charfield")
        self.test_textfield_context = TranslationContext.objects.get(path="test_textfield")

        self.test_content_string = String.objects.get(data="Test content")
        self.more_test_content_string = String.objects.get(data="More test content")

    def test_get_status_display(self):
        self.assertEqual(self.translation.get_status_display(), "Waiting for translations")

    def test_get_status_display_with_translations(self):
        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=self.test_charfield_context,
            locale=self.fr_locale,
            data="Contenu de test",
        )

        StringTranslation.objects.create(
            translation_of=self.more_test_content_string,
            context=self.test_textfield_context,
            locale=self.fr_locale,
            data="Plus de contenu de test",
        )

        self.assertEqual(self.translation.get_status_display(), "Up to date")


class TestSaveTarget(TestCase):
    def setUp(self):
        self.fr_locale = Locale.objects.create(language_code="fr")

        self.page = create_test_page(
            title="Test page",
            slug="test-slug",
            test_charfield="Test content",
            test_textfield="More test content"
        )
        self.source = TranslationSource.objects.get()
        self.translation = Translation.objects.create(
            source=self.source,
            target_locale=self.fr_locale,
        )

        self.test_charfield_context = TranslationContext.objects.get(path="test_charfield")
        self.test_textfield_context = TranslationContext.objects.get(path="test_textfield")

        self.test_content_string = String.objects.get(data="Test content")
        self.more_test_content_string = String.objects.get(data="More test content")

    def test_save_target(self):
        self.translation.save_target()

        # Should create the page with English content
        translated_page = self.page.get_translation(self.fr_locale)
        self.assertEqual(translated_page.test_charfield, "Test content")
        self.assertEqual(translated_page.test_textfield, "More test content")

        self.assertTrue(translated_page.live)

    def test_save_target_as_draft(self):
        self.translation.save_target(publish=False)

        # Should create the page with English content
        translated_page = self.page.get_translation(self.fr_locale)
        self.assertEqual(translated_page.test_charfield, "Test content")
        self.assertEqual(translated_page.test_textfield, "More test content")

        self.assertFalse(translated_page.live)

    def test_save_target_with_translations(self):
        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=self.test_charfield_context,
            locale=self.fr_locale,
            data="Contenu de test",
        )

        StringTranslation.objects.create(
            translation_of=self.more_test_content_string,
            context=self.test_textfield_context,
            locale=self.fr_locale,
            data="Plus de contenu de test",
        )

        self.translation.save_target()

        translated_page = self.page.get_translation(self.fr_locale)
        self.assertEqual(translated_page.test_charfield, "Contenu de test")
        self.assertEqual(translated_page.test_textfield, "Plus de contenu de test")

    def test_save_target_with_partial_translations(self):
        StringTranslation.objects.create(
            translation_of=self.more_test_content_string,
            context=self.test_textfield_context,
            locale=self.fr_locale,
            data="Plus de contenu de test",
        )

        self.translation.save_target()

        translated_page = self.page.get_translation(self.fr_locale)
        self.assertEqual(translated_page.test_charfield, "Test content")
        self.assertEqual(translated_page.test_textfield, "Plus de contenu de test")

    def test_save_target_snippet(self):
        snippet = TestSnippet.objects.create(field="Test content")
        source, created = TranslationSource.get_or_create_from_instance(snippet)
        translation = Translation.objects.create(
            source=source,
            target_locale=self.fr_locale,
        )

        field_context = TranslationContext.objects.get(path="field")
        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=field_context,
            locale=self.fr_locale,
            data="Contenu de test",
        )

        translation.save_target()

        translated_snippet = snippet.get_translation(self.fr_locale)
        self.assertEqual(translated_snippet.field, "Contenu de test")

    def test_save_target_cant_save_snippet_as_draft(self):
        snippet = TestSnippet.objects.create(field="Test content")
        source, created = TranslationSource.get_or_create_from_instance(snippet)
        translation = Translation.objects.create(
            source=source,
            target_locale=self.fr_locale,
        )

        field_context = TranslationContext.objects.get(path="field")
        StringTranslation.objects.create(
            translation_of=self.test_content_string,
            context=field_context,
            locale=self.fr_locale,
            data="Contenu de test",
        )

        with self.assertRaises(CannotSaveDraftError):
            translation.save_target(publish=False)


class TestDeleteSourceDisablesTranslation(TestCase):
    def setUp(self):
        self.fr_locale = Locale.objects.create(language_code="fr")

    def test_page(self):
        page = create_test_page(
            title="Test page",
            slug="test-slug",
            test_charfield="Test content",
            test_textfield="More test content"
        )

        source = TranslationSource.objects.get()
        translation = Translation.objects.create(
            source=source,
            target_locale=self.fr_locale,
        )

        translation.refresh_from_db()
        self.assertTrue(translation.enabled)

        page.delete()

        translation.refresh_from_db()
        self.assertFalse(translation.enabled)

    def test_snippet(self):
        snippet = TestSnippet.objects.create(field="Test content")

        source, created = TranslationSource.get_or_create_from_instance(snippet)
        translation = Translation.objects.create(
            source=source,
            target_locale=self.fr_locale,
        )

        translation.refresh_from_db()
        self.assertTrue(translation.enabled)

        snippet.delete()

        translation.refresh_from_db()
        self.assertFalse(translation.enabled)


class TestDeleteDestinationDisablesTranslation(TestCase):
    def setUp(self):
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.es_locale = Locale.objects.create(language_code="es")

    def test_page(self):
        page = create_test_page(
            title="Test page",
            slug="test-slug",
            test_charfield="Test content",
            test_textfield="More test content"
        )

        source = TranslationSource.objects.get()
        fr_translation = Translation.objects.create(
            source=source,
            target_locale=self.fr_locale,
        )
        fr_translation.save_target()

        # Create a separate spanish translation to make sure that isn't disabled
        es_translation = Translation.objects.create(
            source=source,
            target_locale=self.es_locale,
        )
        es_translation.save_target()

        fr_translation.refresh_from_db()
        es_translation.refresh_from_db()
        self.assertTrue(fr_translation.enabled)
        self.assertTrue(es_translation.enabled)

        page.get_translation(self.fr_locale).delete()

        fr_translation.refresh_from_db()
        es_translation.refresh_from_db()
        self.assertFalse(fr_translation.enabled)
        self.assertTrue(es_translation.enabled)

    def test_snippet(self):
        snippet = TestSnippet.objects.create(field="Test content")

        source, created = TranslationSource.get_or_create_from_instance(snippet)
        fr_translation = Translation.objects.create(
            source=source,
            target_locale=self.fr_locale,
        )
        fr_translation.save_target()

        # Create a separate spanish translation to make sure that isn't disabled
        es_translation = Translation.objects.create(
            source=source,
            target_locale=self.es_locale,
        )
        es_translation.save_target()

        fr_translation.refresh_from_db()
        es_translation.refresh_from_db()
        self.assertTrue(fr_translation.enabled)
        self.assertTrue(es_translation.enabled)

        snippet.get_translation(self.fr_locale).delete()

        fr_translation.refresh_from_db()
        es_translation.refresh_from_db()
        self.assertFalse(fr_translation.enabled)
        self.assertTrue(es_translation.enabled)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class WagtailLocalesAppConfig(AppConfig):
    name = 'wagtail_localize.locales'
    label = 'wagtaillocales'
    verbose_name = _("Wagtail locales (Wagtail Localize version)")



import json
import tempfile
from collections import defaultdict

import polib
from django.contrib.admin.utils import quote
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied, ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, transaction
from django.http import Http404, HttpResponse
from django.urls import reverse
from django.shortcuts import get_object_or_404, render, redirect
from django.utils.functional import cached_property
from django.utils.text import capfirst, slugify
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST
from modelcluster.fields import ParentalKey
from rest_framework import serializers, status
from rest_framework.decorators import api_view
from rest_framework.response import Response
from wagtail.admin import messages
from wagtail.admin.edit_handlers import TabbedInterface, ObjectList, BaseCompositeEditHandler, FieldPanel, PageChooserPanel
from wagtail.admin.navigation import get_explorable_root_page
from wagtail.admin.templatetags.wagtailadmin_tags import avatar_url
from wagtail.admin.views.pages.utils import get_valid_next_url_from_request
from wagtail.core import blocks
from wagtail.core.fields import StreamField
from wagtail.core.models import Page, TranslatableMixin
from wagtail.core.utils import cautious_slugify
from wagtail.documents.blocks import DocumentChooserBlock
from wagtail.documents.models import AbstractDocument
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import AbstractImage
from wagtail.snippets.blocks import SnippetChooserBlock
from wagtail.snippets.models import get_snippet_models
from wagtail.snippets.permissions import get_permission_name, user_can_edit_snippet_type
from wagtail.snippets.views.snippets import get_snippet_edit_handler

from wagtail_localize.compat import DATE_FORMAT
from wagtail_localize.machine_translators import get_machine_translator
from wagtail_localize.models import Translation, StringTranslation, StringSegment, OverridableSegment, SegmentOverride
from wagtail_localize.segments import StringSegmentValue


class UserSerializer(serializers.ModelSerializer):
    full_name = serializers.ReadOnlyField(source='get_full_name')
    avatar_url = serializers.SerializerMethodField('get_avatar_url')

    def get_avatar_url(self, user):
        return avatar_url(user, size=25)

    class Meta:
        model = get_user_model()
        fields = ['full_name', 'avatar_url']


class StringTranslationSerializer(serializers.ModelSerializer):
    string_id = serializers.ReadOnlyField(source='translation_of_id')
    segment_id = serializers.SerializerMethodField('get_segment_id')
    error = serializers.ReadOnlyField(source='get_error')
    comment = serializers.ReadOnlyField(source='get_comment')
    last_translated_by = UserSerializer()

    def get_segment_id(self, translation):
        if 'translation_source' in self.context:
            translation_source = self.context['translation_source']
            return translation_source.stringsegment_set.filter(
                string_id=translation.translation_of_id,
                context_id=translation.context_id,
            ).values_list('id', flat=True).first()

    class Meta:
        model = StringTranslation
        fields = ['string_id', 'segment_id', 'data', 'error', 'comment', 'last_translated_by']


class SegmentOverrideSerializer(serializers.ModelSerializer):
    segment_id = serializers.SerializerMethodField('get_segment_id')
    error = serializers.ReadOnlyField(source='get_error')

    def get_segment_id(self, override):
        if 'translation_source' in self.context:
            translation_source = self.context['translation_source']

            try:
                return translation_source.overridablesegment_set.only('id').get(
                    context_id=override.context_id,
                ).id
            except OverridableSegment.DoesNotExist:
                return

    class Meta:
        model = SegmentOverride
        fields = ['segment_id', 'data', 'error']


class TabHelper:
    def __init__(self, instance):
        self.instance = instance

    @cached_property
    def edit_handler(self):
        if isinstance(self.instance, Page):
            return self.instance.get_edit_handler()
        else:
            return get_snippet_edit_handler(self.instance.__class__)

    @cached_property
    def tabs(self):
        tabs = []

        if isinstance(self.edit_handler, TabbedInterface):
            for tab in self.edit_handler.children:
                # On Pages, the TabbedInterface children are instances of ObjectList
                # which contain the fields
                # On Snippets, the fields can be added directly into the TabbedInterface
                # In this case, we do not want to add any tabs and instead just fall back
                # to the default "Content" tab added below.
                if isinstance(tab, ObjectList):
                    tabs.append(tab.heading)

        # Add a default "Content" tab if this object doesn't have any tabs
        if not tabs:
            tabs = [_("Content")]

        return tabs

    @property
    def tabs_with_slugs(self):
        return [
            {
                'label': label,
                'slug': cautious_slugify(label),
            }
            for label in self.tabs
        ]

    @cached_property
    def field_tab_mapping(self):
        if isinstance(self.edit_handler, TabbedInterface):
            field_tabs = {}
            for tab in self.edit_handler.children:
                for tab_field in tab.required_fields():
                    field_tabs[tab_field] = tab.heading

                for tab_formset in tab.required_formsets().keys():
                    field_tabs[tab_formset] = tab.heading

            return field_tabs
        else:
            return {}

    def get_field_tab(self, field_name):
        if field_name in self.field_tab_mapping:
            return self.field_tab_mapping[field_name]
        else:
            raise KeyError(f"Cannot find tab for field '{field_name}''")

    @cached_property
    def field_ordering_mapping(self):
        if isinstance(self.edit_handler, TabbedInterface):
            field_orderings = {}
            order = 0
            for tab in self.edit_handler.children:
                for tab_field in tab.required_fields():
                    # TODO(someday): Orderings of fields within inline panels.
                    # (currently, they will all be assigned the same order value,
                    # so they will end up being order by how they are defined on
                    # the model instead of the panel definition.
                    # But this should be OK for most people)
                    field_orderings[tab_field] = order
                    order += 1

                for tab_formset in tab.required_formsets().keys():
                    field_orderings[tab_formset] = order
                    order += 1

            return field_orderings
        else:
            return {}

    def get_field_order(self, field_name):
        if field_name in self.field_ordering_mapping:
            return self.field_ordering_mapping[field_name]
        else:
            raise KeyError(f"Cannot find ordering for field '{field_name}''")

    @cached_property
    def field_edit_handler_mapping(self):
        # TODO (someday): Extract mappings out of inline panels
        field_edit_handlers = {}

        def walk(edit_handler):
            if isinstance(edit_handler, BaseCompositeEditHandler):
                for child in edit_handler.children:
                    walk(child)

            elif isinstance(edit_handler, FieldPanel):
                field_edit_handlers[edit_handler.field_name] = edit_handler

        walk(self.edit_handler)

        return field_edit_handlers

    def get_field_edit_handler(self, field_name):
        if field_name in self.field_edit_handler_mapping:
            return self.field_edit_handler_mapping[field_name]


class FieldHasNoEditPanelError(KeyError):
    pass


def get_segment_location_info(source_instance, tab_helper, content_path, widget=False):
    context_path_components = content_path.split('.')
    field = source_instance._meta.get_field(context_path_components[0])

    # Work out which tab the segment is on from edit handler
    try:
        tab = cautious_slugify(tab_helper.get_field_tab(field.name))
    except KeyError:
        raise FieldHasNoEditPanelError

    order = tab_helper.get_field_order(field.name)

    def widget_from_field(field):
        if isinstance(field, models.ForeignKey):
            if issubclass(field.related_model, Page):
                edit_handler = tab_helper.get_field_edit_handler(field.name)

                if isinstance(edit_handler, PageChooserPanel):
                    return {
                        'type': 'page_chooser',
                        'allowed_page_types': [
                            '{app}.{model}'.format(
                                app=model._meta.app_label,
                                model=model._meta.model_name)
                            for model in edit_handler.target_models()
                        ],
                    }

                else:
                    return {
                        'type': 'unknown'
                    }

            elif issubclass(field.related_model, AbstractDocument):
                return {
                    'type': 'document_chooser'
                }

            elif issubclass(field.related_model, AbstractImage):
                return {
                    'type': 'image_chooser'
                }

            elif issubclass(field.related_model, tuple(get_snippet_models())):
                return {
                    'type': 'snippet_chooser',
                    'snippet_model': {
                        'app_label': field.related_model._meta.app_label,
                        'model_name': field.related_model._meta.model_name,
                        'verbose_name': field.related_model._meta.verbose_name,
                        'verbose_name_plural': field.related_model._meta.verbose_name_plural,
                    },
                    'chooser_url': reverse('wagtailsnippets:choose', args=[field.related_model._meta.app_label, field.related_model._meta.model_name])
                }

        elif isinstance(field, (models.CharField, models.TextField, models.EmailField, models.URLField)):
            return {
                'type': 'text',
            }

        return {
            'type': 'unknown'
        }

    def widget_from_block(block):
        if isinstance(block, blocks.PageChooserBlock):
            return {
                'type': 'page_chooser',
                'allowed_page_types': [
                    '{app}.{model}'.format(
                        app=model._meta.app_label,
                        model=model._meta.model_name)

                    # Note: Unlike PageChooserPanel, the block doesn't automatically fall back to [Page]
                    for model in block.target_models or [Page]
                ]
            }

        elif isinstance(block, DocumentChooserBlock):
            return {
                'type': 'document_chooser'
            }

        elif isinstance(block, ImageChooserBlock):
            return {
                'type': 'image_chooser'
            }

        elif isinstance(block, SnippetChooserBlock):
            return {
                'type': 'snippet_chooser',
                'snippet_model': {
                    'app_label': block.target_model._meta.app_label,
                    'model_name': block.target_model._meta.model_name,
                    'verbose_name': block.target_model._meta.verbose_name,
                    'verbose_name_plural': block.target_model._meta.verbose_name_plural,
                },
                'chooser_url': reverse('wagtailsnippets:choose', args=[block.target_model._meta.app_label, block.target_model._meta.model_name])
            }

        elif isinstance(block, (blocks.CharBlock, blocks.TextBlock, blocks.EmailBlock, blocks.URLBlock)):
            return {
                'type': 'text',
            }

        return {
            'type': 'unknown'
        }

    if isinstance(field, StreamField):
        stream_value = field.value_from_object(source_instance)
        stream_blocks_by_id = {
            block.id: block
            for block in field.value_from_object(source_instance)
        }
        block_id = context_path_components[1]
        block_value = stream_blocks_by_id[block_id]
        block_type = stream_value.stream_block.child_blocks[block_value.block_type]

        if isinstance(block_type, blocks.StructBlock):
            block_field_name = context_path_components[2]
            block_field = block_type.child_blocks[block_field_name].label
        else:
            block_field = None

        return {
            'tab': tab,
            'field': capfirst(block_type.label),
            'order': order,
            'blockId': block_id,
            'fieldHelpText': '',
            'subField': block_field,
            'widget': widget_from_block(block_type) if widget else None,
        }

    elif (
        isinstance(field, (models.ManyToOneRel))
        and isinstance(field.remote_field, ParentalKey)
        and issubclass(field.related_model, TranslatableMixin)
    ):
        child_field = field.related_model._meta.get_field(context_path_components[2])

        return {
            'tab': tab,
            'field': capfirst(field.related_model._meta.verbose_name),
            'order': order,
            'blockId': context_path_components[1],
            'fieldHelpText': child_field.help_text,
            'subField': capfirst(child_field.verbose_name),
            'widget': widget_from_field(child_field) if widget else None,
        }

    else:
        return {
            'tab': tab,
            'field': capfirst(field.verbose_name),
            'order': order,
            'blockId': None,
            'fieldHelpText': field.help_text,
            'subField': None,
            'widget': widget_from_field(field) if widget else None,
        }


def edit_translation(request, translation, instance):
    if isinstance(instance, Page):
        # Page
        # Note: Edit permission is already checked by the edit page view

        page_perms = instance.permissions_for_user(request.user)

        is_live = instance.live
        is_locked = instance.locked

        if instance.live_revision:
            last_published_at = instance.live_revision.created_at
            last_published_by = instance.live_revision.user
        else:
            last_published_at = instance.last_published_at
            last_published_by = None

        if instance.live:
            live_url = instance.full_url
        else:
            live_url = None

        can_publish = page_perms.can_publish()
        can_unpublish = page_perms.can_unpublish()
        can_lock = page_perms.can_lock()
        can_unlock = page_perms.can_unlock()
        can_delete = page_perms.can_delete()

    else:
        # Snippet
        # Note: Edit permission is already checked by the edit snippet view

        is_live = True
        is_locked = False
        last_published_at = None
        last_published_by = None
        live_url = None

        can_publish = True
        can_unpublish = False
        can_lock = False
        can_unlock = False
        can_delete = request.user.has_perm(get_permission_name('delete', instance.__class__))

    source_instance = translation.source.get_source_instance()

    if request.method == 'POST':
        if request.POST.get('action') == 'publish':
            if isinstance(instance, Page):
                if not page_perms.can_publish():
                    raise PermissionDenied

            try:
                translation.save_target(user=request.user, publish=True)

            except ValidationError:
                messages.error(request, _("New validation errors were found when publishing '{object}' in {locale}. Please fix them or click publish again to ignore these translations for now.").format(
                    object=str(instance),
                    locale=translation.target_locale.get_display_name()
                ))

            else:
                # Refresh instance to title in success message is up to date
                instance.refresh_from_db()

                string_segments = translation.source.stringsegment_set.all().order_by('order')
                string_translations = string_segments.get_translations(translation.target_locale)

                # Using annotate_translation as this ignores errors by default (so both errors and missing segments treated the same)
                if string_segments.annotate_translation(translation.target_locale).filter(translation__isnull=True).exists():
                    # One or more strings had an error
                    messages.warning(request, _("Published '{object}' in {locale} with missing translations - see below.").format(
                        object=str(instance),
                        locale=translation.target_locale.get_display_name()
                    ))

                else:
                    messages.success(request, _("Published '{object}' in {locale}.").format(
                        object=str(instance),
                        locale=translation.target_locale.get_display_name()
                    ))

        return redirect(request.path)

    string_segments = translation.source.stringsegment_set.all().order_by('order')
    string_translations = string_segments.get_translations(translation.target_locale)

    overridable_segments = translation.source.overridablesegment_set.all().order_by('order')
    segment_overrides = overridable_segments.get_overrides(translation.target_locale)
    related_object_segments = translation.source.relatedobjectsegment_set.all().order_by('order')

    tab_helper = TabHelper(source_instance)

    breadcrumb = []
    title_segment_id = None
    if isinstance(instance, Page):
        # find the closest common ancestor of the pages that this user has direct explore permission
        # (i.e. add/edit/publish/lock) over; this will be the root of the breadcrumb
        cca = get_explorable_root_page(request.user)
        if cca:
            breadcrumb = [
                {
                    'id': page.id,
                    'isRoot': page.is_root(),
                    'title': page.title,
                    'exploreUrl': reverse('wagtailadmin_explore_root') if page.is_root() else reverse('wagtailadmin_explore', args=[page.id]),
                }
                for page in instance.get_ancestors(inclusive=False).descendant_of(cca, inclusive=True)
            ]

        # Set to the ID of a string segment that represents the title.
        # If this segment has a translation, the title will be replaced with that translation.
        try:
            title_segment_id = string_segments.get(context__path='title').id
        except StringSegment.DoesNotExist:
            pass

    machine_translator = None
    translator = get_machine_translator()
    if translator and translator.can_translate(translation.source.locale, translation.target_locale):
        machine_translator = {
            'name': translator.display_name,
            'url': reverse('wagtail_localize:machine_translate', args=[translation.id]),
        }

    segments = []

    for segment in string_segments:
        try:
            location_info = get_segment_location_info(source_instance, tab_helper, segment.context.path)
        except FieldHasNoEditPanelError:
            continue

        segments.append({
            'type': 'string',
            'id': segment.id,
            'contentPath': segment.context.path,
            'source': segment.string.data,
            'location': location_info,
            'editUrl': reverse('wagtail_localize:edit_string_translation', kwargs={'translation_id': translation.id, 'string_segment_id': segment.id}),
            'order': segment.order,
        })

    for segment in overridable_segments:
        try:
            location_info = get_segment_location_info(source_instance, tab_helper, segment.context.path, widget=True)
        except FieldHasNoEditPanelError:
            continue

        segments.append({
            'type': 'synchronised_value',
            'id': segment.id,
            'contentPath': segment.context.path,
            'location': location_info,
            'value': segment.data,
            'editUrl': reverse('wagtail_localize:edit_override', kwargs={'translation_id': translation.id, 'overridable_segment_id': segment.id}),
            'order': segment.order,
        })

    def get_source_object_info(segment):
        instance = segment.get_source_instance()

        if isinstance(instance, Page):
            return {
                'title': str(instance),
                'isLive': instance.live,
                'liveUrl': instance.full_url,
                'editUrl': reverse('wagtailadmin_pages:edit', args=[instance.id]),
                'createTranslationRequestUrl': reverse('wagtail_localize:submit_page_translation', args=[instance.id]),
            }

        else:
            return {
                'title': str(instance),
                'isLive': True,
                'editUrl': reverse('wagtailsnippets:edit', args=[instance._meta.app_label, instance._meta.model_name, quote(instance.id)]),
                'createTranslationRequestUrl': reverse('wagtail_localize:submit_snippet_translation', args=[instance._meta.app_label, instance._meta.model_name, quote(instance.id)]),
            }

    def get_dest_object_info(segment):
        instance = segment.object.get_instance_or_none(translation.target_locale)
        if not instance:
            return

        if isinstance(instance, Page):
            return {
                'title': str(instance),
                'isLive': instance.live,
                'liveUrl': instance.full_url,
                'editUrl': reverse('wagtailadmin_pages:edit', args=[instance.id]),
            }

        else:
            return {
                'title': str(instance),
                'isLive': True,
                'editUrl': reverse('wagtailsnippets:edit', args=[instance._meta.app_label, instance._meta.model_name, quote(instance.id)]),
            }

    def get_translation_progress(segment, locale):
        try:
            translation = Translation.objects.get(source__object_id=segment.object_id, target_locale=locale, enabled=True)

        except Translation.DoesNotExist:
            return None

        total_segments, translated_segments = translation.get_progress()

        return {
            'totalSegments': total_segments,
            'translatedSegments': translated_segments,
        }

    for segment in related_object_segments:
        try:
            location_info = get_segment_location_info(source_instance, tab_helper, segment.context.path)
        except FieldHasNoEditPanelError:
            continue

        segments.append({
            'type': 'related_object',
            'id': segment.id,
            'contentPath': segment.context.path,
            'location': location_info,
            'order': segment.order,
            'source': get_source_object_info(segment),
            'dest': get_dest_object_info(segment),
            'translationProgress': get_translation_progress(segment, translation.target_locale),
        })

    # Order segments by how they appear in the content panels
    # segment['location']['order'] is the content panel ordering
    # segment['order'] is the model field ordering
    # User's expect segments to follow the panel ordering as that's the ordering
    # that is used in the page editor of the source page. However, segments that
    # come from the same streamfield/inline panel are given the same value for
    # panel ordering, so we need to order by model field ordering as well (all
    # segments have a unique value for model field ordering)
    segments.sort(key=lambda segment: (segment['location']['order'], segment['order']))

    # Display a warning to the user if the schema of the source model has been updated since the source was last updated
    if translation.source.schema_out_of_date():
        messages.warning(
            request,
            _(
                "The data model for '{model_name}' has been changed since the last translation sync. "
                "If any new fields have been added recently, these may not be visible until the next translation sync."
            ).format(model_name=capfirst(source_instance._meta.verbose_name))
        )

    return render(request, 'wagtail_localize/admin/edit_translation.html', {
        'translation': translation,

        # These props are passed directly to the TranslationEditor react component
        'props': json.dumps({
            'adminBaseUrl': reverse('wagtailadmin_home'),
            'object': {
                'title': str(instance),
                'titleSegmentId': title_segment_id,
                'isLive': is_live,
                'isLocked': is_locked,
                'lastPublishedDate': last_published_at.strftime(DATE_FORMAT) if last_published_at is not None else None,
                'lastPublishedBy': UserSerializer(last_published_by).data if last_published_by is not None else None,
                'liveUrl': live_url,
            },
            'breadcrumb': breadcrumb,
            'tabs': tab_helper.tabs_with_slugs,
            'sourceLocale': {
                'code': translation.source.locale.language_code,
                'displayName': translation.source.locale.get_display_name(),
            },
            'locale': {
                'code': translation.target_locale.language_code,
                'displayName': translation.target_locale.get_display_name(),
            },
            'translations': [
                {
                    'title': str(translated_instance),
                    'locale': {
                        'code': translated_instance.locale.language_code,
                        'displayName': translated_instance.locale.get_display_name(),
                    },
                    'editUrl': reverse('wagtailadmin_pages:edit', args=[translated_instance.id]) if isinstance(translated_instance, Page) else reverse('wagtailsnippets:edit', args=[translated_instance._meta.app_label, translated_instance._meta.model_name, quote(translated_instance.id)]),
                }
                for translated_instance in instance.get_translations().select_related('locale')
            ],
            'perms': {
                'canPublish': can_publish,
                'canUnpublish': can_unpublish,
                'canLock': can_lock,
                'canUnlock': can_unlock,
                'canDelete': can_delete,
            },
            'links': {
                'downloadPofile': reverse('wagtail_localize:download_pofile', args=[translation.id]),
                'uploadPofile': reverse('wagtail_localize:upload_pofile', args=[translation.id]),
                'unpublishUrl': reverse('wagtailadmin_pages:unpublish', args=[instance.id]) if isinstance(instance, Page) else None,
                'lockUrl': reverse('wagtailadmin_pages:lock', args=[instance.id]) if isinstance(instance, Page) else None,
                'unlockUrl': reverse('wagtailadmin_pages:unlock', args=[instance.id]) if isinstance(instance, Page) else None,
                'deleteUrl': reverse('wagtailadmin_pages:delete', args=[instance.id]) if isinstance(instance, Page) else reverse('wagtailsnippets:delete', args=[instance._meta.app_label, instance._meta.model_name, quote(instance.pk)]),
                'stopTranslationUrl': reverse('wagtail_localize:stop_translation', args=[translation.id]),
            },
            'previewModes': [
                {
                    'mode': mode,
                    'label': label,
                    'url': reverse('wagtail_localize:preview_translation', args=[translation.id]) if mode == instance.default_preview_mode else reverse('wagtail_localize:preview_translation', args=[translation.id, mode]),
                }
                for mode, label in (instance.preview_modes if isinstance(instance, Page) else [])
            ],
            'machineTranslator': machine_translator,
            'segments': segments,

            # We serialize the translation data using Django REST Framework.
            # This gives us a consistent representation with the APIs so we
            # can dynamically update translations in the view.
            'initialStringTranslations': StringTranslationSerializer(string_translations, many=True, context={'translation_source': translation.source}).data,
            'initialOverrides': SegmentOverrideSerializer(segment_overrides, many=True, context={'translation_source': translation.source}).data,
        }, cls=DjangoJSONEncoder)
    })


def user_can_edit_instance(user, instance):
    if isinstance(instance, Page):
        # Page
        page_perms = instance.permissions_for_user(user)
        return page_perms.can_edit()

    else:
        # Snippet
        return user_can_edit_snippet_type(user, instance.__class__)


def preview_translation(request, translation_id, mode=None):
    translation = get_object_or_404(Translation, id=translation_id)

    instance = translation.get_target_instance()

    if not isinstance(instance, Page):
        raise Http404

    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    if mode is None:
        mode = instance.default_preview_mode

    if mode not in dict(instance.preview_modes):
        raise Http404

    translation = translation.source.get_ephemeral_translated_instance(translation.target_locale, fallback=True)

    return translation.make_preview_request(request, mode)


@require_POST
def stop_translation(request, translation_id):
    translation = get_object_or_404(Translation, id=translation_id)

    instance = translation.get_target_instance()
    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    translation.enabled = False
    translation.save(update_fields=['enabled'])

    next_url = get_valid_next_url_from_request(request)
    if not next_url:
        # Note: You should always provide a next URL when using this view!
        next_url = reverse('wagtailadmin_home')

    messages.success(
        request,
        _("Translation has been stopped.")
    )

    return redirect(next_url)


@require_POST
def restart_translation(request, translation, instance):
    # This view is hooked in using the before_edit_page hook so we don't need to check for edit permission
    translation.enabled = True
    translation.save(update_fields=['enabled'])

    messages.success(
        request,
        _("Translation has been restarted.")
    )

    if isinstance(instance, Page):
        return redirect('wagtailadmin_pages:edit', instance.id)
    else:
        return redirect('wagtailsnippets:edit', instance._meta.app_label, instance._meta.model_name, quote(instance.pk))


@api_view(['PUT', 'DELETE'])
def edit_string_translation(request, translation_id, string_segment_id):
    translation = get_object_or_404(Translation, id=translation_id)
    string_segment = get_object_or_404(StringSegment, id=string_segment_id)

    if string_segment.context.object_id != translation.source.object_id:
        raise Http404

    instance = translation.get_target_instance()
    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    if request.method == 'PUT':
        string_translation, created = StringTranslation.objects.update_or_create(
            translation_of_id=string_segment.string_id,
            locale_id=translation.target_locale_id,
            context_id=string_segment.context_id,
            defaults={
                'data': request.POST['value'],
                'translation_type': StringTranslation.TRANSLATION_TYPE_MANUAL,
                'tool_name': "",
                'last_translated_by': request.user,
                'has_error': False,
                'field_error': "",
            }
        )

        return Response(StringTranslationSerializer(string_translation, context={'translation_source': translation.source}).data, status=status.HTTP_201_CREATED if created else status.HTTP_200_OK)

    elif request.method == 'DELETE':
        string_translation = StringTranslation.objects.filter(
            translation_of_id=string_segment.string_id,
            locale_id=translation.target_locale_id,
            context_id=string_segment.context_id,
        ).first()

        if string_translation:
            string_translation.delete()

            return Response(StringTranslationSerializer(string_translation, context={'translation_source': translation.source}).data, status=status.HTTP_200_OK)

        else:
            # Note: this is still considered a success in the frontend
            return Response(status=status.HTTP_404_NOT_FOUND)


@api_view(['PUT', 'DELETE'])
def edit_override(request, translation_id, overridable_segment_id):
    translation = get_object_or_404(Translation, id=translation_id)
    overridable_segment = get_object_or_404(OverridableSegment, id=overridable_segment_id)

    if overridable_segment.context.object_id != translation.source.object_id:
        raise Http404

    instance = translation.get_target_instance()
    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    if request.method == 'PUT':
        override, created = SegmentOverride.objects.update_or_create(
            locale_id=translation.target_locale_id,
            context_id=overridable_segment.context_id,
            defaults={
                'data_json': json.dumps(request.POST['value']),
                'last_translated_by': request.user,
            }
        )

        return Response(SegmentOverrideSerializer(override, context={'translation_source': translation.source}).data, status=status.HTTP_201_CREATED if created else status.HTTP_200_OK)

    elif request.method == 'DELETE':
        override = SegmentOverride.objects.filter(
            locale_id=translation.target_locale_id,
            context_id=overridable_segment.context_id,
        ).first()

        if override:
            override.delete()

            return Response(SegmentOverrideSerializer(override, context={'translation_source': translation.source}).data, status=status.HTTP_200_OK)

        else:
            # Note: this is still considered a success in the frontend
            return Response(status=status.HTTP_404_NOT_FOUND)


def download_pofile(request, translation_id):
    translation = get_object_or_404(Translation, id=translation_id)

    instance = translation.get_target_instance()
    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    response = HttpResponse(str(translation.export_po()), content_type="text/x-gettext-translation")
    response["Content-Disposition"] = (
        "attachment; filename=%s-%s.po" % (
            slugify(translation.source.object_repr),
            translation.target_locale.language_code,
        )
    )
    return response


@require_POST
def upload_pofile(request, translation_id):
    translation = get_object_or_404(Translation, id=translation_id)

    instance = translation.get_target_instance()
    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    do_import = True

    with tempfile.NamedTemporaryFile() as f:
        # Note: polib.pofile accepts either a filename or contents. We cannot pass the
        # contents directly into polib.pofile or users could upload a file containing
        # a filename and this will be read by polib!
        f.write(request.FILES["file"].read())
        f.flush()

        try:
            po = polib.pofile(f.name)

        except (OSError, UnicodeDecodeError):
            # Annoyingly, POLib uses OSError for parser exceptions...
            messages.error(
                request,
                _("Please upload a valid PO file.")
            )
            do_import = False

    if do_import:
        translation_id = po.metadata['X-WagtailLocalize-TranslationID']
        if translation_id != str(translation.uuid):
            messages.error(
                request,
                _("Cannot import PO file that was created for a different translation.")
            )
            do_import = False

    if do_import:
        translation.import_po(po, user=request.user, tool_name="PO File")

        messages.success(
            request,
            _("Successfully imported translations from PO File.")
        )

    # Work out where to redirect to
    next_url = get_valid_next_url_from_request(request)
    if not next_url:
        # Note: You should always provide a next URL when using this view!
        next_url = reverse('wagtailadmin_home')

    return redirect(next_url)


@require_POST
def machine_translate(request, translation_id):
    translation = get_object_or_404(Translation, id=translation_id)

    instance = translation.get_target_instance()
    if not user_can_edit_instance(request.user, instance):
        raise PermissionDenied

    translator = get_machine_translator()
    if translator is None:
        raise Http404

    if not translator.can_translate(translation.source.locale, translation.target_locale):
        raise Http404

    # Get segments
    segments = defaultdict(list)
    for string_segment in translation.source.stringsegment_set.all().select_related("context", "string"):
        segment = StringSegmentValue(
            string_segment.context.path, string_segment.string.as_value()
        ).with_order(string_segment.order)
        if string_segment.attrs:
            segment.attrs = json.loads(string_segment.attrs)

        # Don't translate if there already is a translation
        if StringTranslation.objects.filter(
            translation_of_id=string_segment.string_id,
            locale=translation.target_locale,
            context_id=string_segment.context_id,
        ).exists():
            continue

        segments[segment.string].append((string_segment.string_id, string_segment.context_id))

    if segments:
        translations = translator.translate(translation.source.locale, translation.target_locale, segments.keys())

        with transaction.atomic():
            for string, contexts in segments.items():
                for string_id, context_id in contexts:
                    StringTranslation.objects.get_or_create(
                        translation_of_id=string_id,
                        locale=translation.target_locale,
                        context_id=context_id,
                        defaults={
                            'data': translations[string].data,
                            'translation_type': StringTranslation.TRANSLATION_TYPE_MACHINE,
                            'tool_name': translator.display_name,
                            'last_translated_by': request.user,
                            'has_error': False,
                            'field_error': "",
                        }
                    )

        messages.success(
            request,
            _("Successfully translated with {}.").format(translator.display_name)
        )

    else:
        messages.warning(
            request,
            _("There isn't anything left to translate.")
        )

    # Work out where to redirect to
    next_url = get_valid_next_url_from_request(request)
    if not next_url:
        # Note: You should always provide a next URL when using this view!
        next_url = reverse('wagtailadmin_home')

    return redirect(next_url)


def edit_translatable_alias_page(request, page):
    return render(request, 'wagtail_localize/admin/edit_translatable_alias.html', {
        'page': page,
        'page_for_status': page,
        'content_type': page.cached_content_type,
        'next': get_valid_next_url_from_request(request),
        'locale': page.locale,
        'translations': [
            {
                'locale': translation.locale,
                'url': reverse('wagtailadmin_pages:edit', args=[translation.id]),
            }
            for translation in page.get_translations().only('id', 'locale').select_related('locale')
        ],
    })



from django.contrib.messages import get_messages
from django.test import TestCase, override_settings
from django.urls import reverse

from wagtail.core.models import Locale
from wagtail.tests.utils import WagtailTestUtils

from wagtail_localize.models import LocaleSynchronization
from wagtail_localize.locales.components import LOCALE_COMPONENTS


@override_settings(WAGTAIL_CONTENT_LANGUAGES=[("en", "English"), ("fr", "French")])
class TestLocaleIndexView(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

    def get(self, params={}):
        return self.client.get(reverse('wagtaillocales:index'), params)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtaillocales/index.html')


@override_settings(WAGTAIL_CONTENT_LANGUAGES=[("en", "English"), ("fr", "French")])
class TestLocaleCreateView(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()
        self.english = Locale.objects.get()

    def get(self, params={}):
        return self.client.get(reverse('wagtaillocales:add'), params)

    def post(self, post_data={}):
        return self.client.post(reverse('wagtaillocales:add'), post_data)

    def test_default_language(self):
        # we should have loaded with a single locale
        self.assertEqual(self.english.language_code, 'en')
        self.assertEqual(self.english.get_display_name(), "English")

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtaillocales/create.html')

        self.assertEqual(response.context['form'].fields['language_code'].choices, [
            ('fr', 'French')
        ])

    def test_create(self):
        response = self.post({
            'language_code': "fr",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': self.english.id,
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtaillocales:index'))

        # Check that the locale was created
        self.assertTrue(Locale.objects.filter(language_code='fr').exists())

        # Check the sync_from was set
        self.assertTrue(LocaleSynchronization.objects.filter(locale__language_code='fr', sync_from__language_code='en').exists())

    def test_duplicate_not_allowed(self):
        response = self.post({
            'language_code': "en",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': self.english.id,
        })

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'language_code', ['Select a valid choice. en is not one of the available choices.'])

    def test_language_code_must_be_in_settings(self):
        response = self.post({
            'language_code': "ja",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': self.english.id,
        })

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'language_code', ['Select a valid choice. ja is not one of the available choices.'])

    def test_sync_from_required_when_enabled(self):
        response = self.post({
            'language_code': "fr",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': '',
        })

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'component_form', 'sync_from', ['This field is required.'])

        # Check that the locale was not created
        self.assertFalse(Locale.objects.filter(language_code='fr').exists())

    def test_sync_from_not_required_when_disabled(self):
        response = self.post({
            'language_code': "fr",
            'component-wagtail_localize_localesynchronization-enabled': '',
            'component-wagtail_localize_localesynchronization-sync_from': '',
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtaillocales:index'))

        # Check that the locale was created
        self.assertTrue(Locale.objects.filter(language_code='fr').exists())

        # Check the sync_from was not set
        self.assertFalse(LocaleSynchronization.objects.exists())

    def test_sync_from_required_when_component_required(self):
        LOCALE_COMPONENTS[0]['required'] = True
        try:
            response = self.post({
                'language_code': "fr",
                'component-wagtail_localize_localesynchronization-enabled': '',
                'component-wagtail_localize_localesynchronization-sync_from': '',
            })
        finally:
            LOCALE_COMPONENTS[0]['required'] = False

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'component_form', 'sync_from', ['This field is required.'])

        # Check that the locale was not created
        self.assertFalse(Locale.objects.filter(language_code='fr').exists())


@override_settings(WAGTAIL_CONTENT_LANGUAGES=[("en", "English"), ("fr", "French")])
class TestLocaleEditView(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()
        self.english = Locale.objects.get()

    def get(self, params=None, locale=None):
        locale = locale or self.english
        return self.client.get(reverse('wagtaillocales:edit', args=[locale.id]), params or {})

    def post(self, post_data=None, locale=None):
        post_data = post_data or {}
        locale = locale or self.english
        post_data.setdefault('language_code', locale.language_code)
        return self.client.post(reverse('wagtaillocales:edit', args=[locale.id]), post_data)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtaillocales/edit.html')

        self.assertEqual(response.context['form'].fields['language_code'].choices, [
            ('en', 'English'),  # Note: Current value is displayed even though it's in use
            ('fr', 'French')
        ])

    def test_invalid_language(self):
        invalid = Locale.objects.create(language_code='foo')

        response = self.get(locale=invalid)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtaillocales/edit.html')

        self.assertEqual(response.context['form'].fields['language_code'].choices, [
            (None, 'Select a new language'),  # This is shown instead of the current value if invalid
            ('fr', 'French')
        ])

    def test_edit(self):
        response = self.post({
            'language_code': 'fr',
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': self.english.id,
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtaillocales:index'))

        # Check that the locale was edited
        self.english.refresh_from_db()
        self.assertEqual(self.english.language_code, 'fr')

    def test_edit_duplicate_not_allowed(self):
        french = Locale.objects.create(language_code='fr')

        response = self.post({
            'language_code': "en",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': self.english.id,
        }, locale=french)

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'language_code', ['Select a valid choice. en is not one of the available choices.'])

    def test_edit_language_code_must_be_in_settings(self):
        response = self.post({
            'language_code': "ja",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': self.english.id,
        })

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'language_code', ['Select a valid choice. ja is not one of the available choices.'])

    def test_sync_from_required_when_enabled(self):
        response = self.post({
            'language_code': "fr",
            'component-wagtail_localize_localesynchronization-enabled': 'on',
            'component-wagtail_localize_localesynchronization-sync_from': '',
        })

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'component_form', 'sync_from', ['This field is required.'])

    def test_sync_from_not_required_when_disabled(self):
        response = self.post({
            'language_code': "fr",
            'component-wagtail_localize_localesynchronization-enabled': '',
            'component-wagtail_localize_localesynchronization-sync_from': '',
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtaillocales:index'))

        # Check that the locale was edited
        self.english.refresh_from_db()
        self.assertEqual(self.english.language_code, 'fr')

    def test_sync_from_required_when_component_required(self):
        LOCALE_COMPONENTS[0]['required'] = True
        try:
            response = self.post({
                'language_code': "fr",
                'component-wagtail_localize_localesynchronization-enabled': '',
                'component-wagtail_localize_localesynchronization-sync_from': '',
            })
        finally:
            LOCALE_COMPONENTS[0]['required'] = False

        # Should return the form with errors
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'component_form', 'sync_from', ['This field is required.'])


@override_settings(WAGTAIL_CONTENT_LANGUAGES=[("en", "English"), ("fr", "French")])
class TestLocaleDeleteView(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()
        self.english = Locale.objects.get()

    def get(self, params={}, locale=None):
        locale = locale or self.english
        return self.client.get(reverse('wagtaillocales:delete', args=[locale.id]), params)

    def post(self, post_data={}, locale=None):
        locale = locale or self.english
        return self.client.post(reverse('wagtaillocales:delete', args=[locale.id]), post_data)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/generic/confirm_delete.html')

    def test_delete_locale(self):
        french = Locale.objects.create(language_code='fr')

        response = self.post(locale=french)

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtaillocales:index'))

        # Check that the locale was deleted
        self.assertFalse(Locale.objects.filter(language_code='fr').exists())

    def test_cannot_delete_locales_with_pages(self):
        response = self.post()

        self.assertEqual(response.status_code, 200)

        # Check error message
        messages = list(get_messages(response.wsgi_request))
        self.assertEqual(messages[0].level_tag, 'error')
        self.assertEqual(messages[0].message, "This locale cannot be deleted because there are pages and/or other objects using it.\n\n\n\n\n")

        # Check that the locale was not deleted
        self.assertTrue(Locale.objects.filter(language_code='en').exists())



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase, override_settings
from django.urls import reverse
from wagtail.core.models import Page, Locale
from wagtail.tests.utils import WagtailTestUtils

from wagtail_localize.models import Translation, TranslationSource
from wagtail_localize.test.models import TestPage, TestSnippet


def make_test_page(parent, cls=None, **kwargs):
    cls = cls or TestPage
    kwargs.setdefault("title", "Test page")
    return parent.add_child(instance=cls(**kwargs))


@override_settings(
    LANGUAGES=[
        ('en', "English"),
        ('fr', "French"),
        ('de', "German"),
        ('es', "Spanish"),
    ],
    WAGTAIL_CONTENT_LANGUAGES=[
        ('en', "English"),
        ('fr', "French"),
        ('de', "German"),
        ('es', "Spanish"),
    ],
)
class TestTranslationReport(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.en_locale = Locale.objects.get()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        self.en_snippet = TestSnippet.objects.create(field="Test snippet")

        self.en_homepage = Page.objects.get(depth=2)
        self.fr_homepage = self.en_homepage.copy_for_translation(self.fr_locale)
        self.de_homepage = self.fr_homepage.copy_for_translation(self.de_locale)

        self.en_blog_index = make_test_page(self.en_homepage, title="Blog", slug="blog")
        self.en_blog_post = make_test_page(
            self.en_blog_index, title="Blog post", slug="blog-post"
        )

        self.snippet_translation = Translation.objects.create(
            source=TranslationSource.get_or_create_from_instance(self.en_snippet)[0],
            target_locale=self.fr_locale
        )

        self.homepage_translation = Translation.objects.create(
            source=TranslationSource.get_or_create_from_instance(self.en_homepage)[0],
            target_locale=self.fr_locale
        )

        self.de_homepage_translation = Translation.objects.create(
            source=TranslationSource.get_or_create_from_instance(self.fr_homepage)[0],
            target_locale=self.de_locale
        )

        self.blog_index_translation = Translation.objects.create(
            source=TranslationSource.get_or_create_from_instance(self.en_blog_index)[0],
            target_locale=self.fr_locale
        )

        self.blog_post_translation = Translation.objects.create(
            source=TranslationSource.get_or_create_from_instance(self.en_blog_post)[0],
            target_locale=self.fr_locale
        )

    def test_get_empty_report(self):
        Translation.objects.all().delete()

        response = self.client.get(reverse('wagtail_localize:translations_report'))

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "No translations found.")

    def test_get_report(self):
        response = self.client.get(reverse('wagtail_localize:translations_report'))

        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "No translations found.")

        self.assertIn(self.snippet_translation, response.context['object_list'])
        self.assertIn(self.homepage_translation, response.context['object_list'])
        self.assertIn(self.de_homepage_translation, response.context['object_list'])
        self.assertIn(self.blog_index_translation, response.context['object_list'])
        self.assertIn(self.blog_post_translation, response.context['object_list'])

    def test_filter_by_content_type(self):
        snippet_content_type = ContentType.objects.get_for_model(TestSnippet)
        response = self.client.get(reverse('wagtail_localize:translations_report') + '?content_type=' + str(snippet_content_type.id))

        self.assertIn(self.snippet_translation, response.context['object_list'])
        self.assertNotIn(self.homepage_translation, response.context['object_list'])
        self.assertNotIn(self.blog_index_translation, response.context['object_list'])
        self.assertNotIn(self.blog_post_translation, response.context['object_list'])

    def test_filter_by_base_content_type(self):
        snippet_content_type = ContentType.objects.get_for_model(Page)
        response = self.client.get(reverse('wagtail_localize:translations_report') + '?content_type=' + str(snippet_content_type.id))

        self.assertNotIn(self.snippet_translation, response.context['object_list'])
        self.assertIn(self.homepage_translation, response.context['object_list'])
        self.assertIn(self.blog_index_translation, response.context['object_list'])
        self.assertIn(self.blog_post_translation, response.context['object_list'])

    def test_filter_by_source_title(self):
        response = self.client.get(reverse('wagtail_localize:translations_report') + '?source_title=Blog')

        self.assertNotIn(self.snippet_translation, response.context['object_list'])
        self.assertNotIn(self.homepage_translation, response.context['object_list'])
        self.assertIn(self.blog_index_translation, response.context['object_list'])
        self.assertIn(self.blog_post_translation, response.context['object_list'])

    def test_filter_by_source_locale(self):
        response = self.client.get(reverse('wagtail_localize:translations_report') + '?source_locale=fr')

        self.assertNotIn(self.snippet_translation, response.context['object_list'])
        self.assertNotIn(self.homepage_translation, response.context['object_list'])
        self.assertIn(self.de_homepage_translation, response.context['object_list'])
        self.assertNotIn(self.blog_index_translation, response.context['object_list'])
        self.assertNotIn(self.blog_post_translation, response.context['object_list'])

    def test_filter_by_target_locale(self):
        response = self.client.get(reverse('wagtail_localize:translations_report') + '?target_locale=de')

        self.assertNotIn(self.snippet_translation, response.context['object_list'])
        self.assertNotIn(self.homepage_translation, response.context['object_list'])
        self.assertIn(self.de_homepage_translation, response.context['object_list'])
        self.assertNotIn(self.blog_index_translation, response.context['object_list'])
        self.assertNotIn(self.blog_post_translation, response.context['object_list'])



#!/usr/bin/env python

from os import path

from setuptools import find_packages, setup

from wagtail_localize import __version__


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="wagtail-localize",
    version=__version__,
    description="Translation plugin for Wagtail CMS",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Karl Hobley",
    author_email="karl@torchbox.com",
    url="",
    packages=find_packages(),
    include_package_data=True,
    license="BSD",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Wagtail",
        "Framework :: Wagtail :: 2",
    ],
    install_requires=["Django>=2.2,<3.2", "Wagtail>=2.11,<3", "polib>=1.1,<2.0"],
    extras_require={
        "testing": ["dj-database-url==0.5.0", "freezegun==0.3.15"],
        "documentation": [
            "mkdocs==1.1.2",
            "mkdocs-material==6.2.8",
            "mkdocs-mermaid2-plugin==0.5.1",
            "mkdocstrings==0.14.0",
            "mkdocs-include-markdown-plugin==2.8.0",
        ],
    },
    zip_safe=False,
)



# Generated by Django 3.1.3 on 2020-12-01 19:24

from django.db import migrations, models
import django.db.models.deletion
import wagtail.core.blocks
import wagtail.core.fields
import wagtail.documents.blocks
import wagtail.embeds.blocks
import wagtail.images.blocks
import wagtail.snippets.blocks
import wagtail_localize.test.models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0059_apply_collection_ordering'),
        ('wagtail_localize_test', '0011_testmodelwithinvalidforeignkey'),
    ]

    operations = [
        migrations.AddField(
            model_name='testpage',
            name='test_page',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.page'),
        ),
        migrations.AddField(
            model_name='testpage',
            name='test_page_specific_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtail_localize_test.testhomepage'),
        ),
        migrations.AddField(
            model_name='testpage',
            name='test_page_with_restricted_types',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.page'),
        ),
        migrations.AlterField(
            model_name='testgeneratetranslatablefieldspage',
            name='test_streamfield',
            field=wagtail.core.fields.StreamField([('test_charblock', wagtail.core.blocks.CharBlock(max_length=255)), ('test_textblock', wagtail.core.blocks.TextBlock(label='text block')), ('test_emailblock', wagtail.core.blocks.EmailBlock()), ('test_urlblock', wagtail.core.blocks.URLBlock()), ('test_richtextblock', wagtail.core.blocks.RichTextBlock()), ('test_rawhtmlblock', wagtail.core.blocks.RawHTMLBlock()), ('test_blockquoteblock', wagtail.core.blocks.BlockQuoteBlock()), ('test_structblock', wagtail.core.blocks.StructBlock([('field_a', wagtail.core.blocks.TextBlock()), ('field_b', wagtail.core.blocks.TextBlock())])), ('test_listblock', wagtail.core.blocks.ListBlock(wagtail.core.blocks.TextBlock())), ('test_nestedstreamblock', wagtail.core.blocks.StreamBlock([('block_a', wagtail.core.blocks.TextBlock()), ('block_b', wagtail.core.blocks.TextBlock())])), ('test_customstructblock', wagtail.core.blocks.StructBlock([('field_a', wagtail.core.blocks.TextBlock()), ('field_b', wagtail.core.blocks.TextBlock())])), ('test_customblockwithoutextractmethod', wagtail_localize.test.models.CustomBlockWithoutExtractMethod()), ('test_pagechooserblock', wagtail.core.blocks.PageChooserBlock()), ('test_pagechooserblock_with_restricted_types', wagtail.core.blocks.PageChooserBlock(['wagtail_localize_test.TestHomePage', 'wagtail_localize_test.TestPage'])), ('test_imagechooserblock', wagtail.images.blocks.ImageChooserBlock()), ('test_documentchooserblock', wagtail.documents.blocks.DocumentChooserBlock()), ('test_snippetchooserblock', wagtail.snippets.blocks.SnippetChooserBlock(wagtail_localize.test.models.TestSnippet)), ('test_embedblock', wagtail.embeds.blocks.EmbedBlock())], blank=True),
        ),
        migrations.AlterField(
            model_name='testpage',
            name='test_streamfield',
            field=wagtail.core.fields.StreamField([('test_charblock', wagtail.core.blocks.CharBlock(max_length=255)), ('test_textblock', wagtail.core.blocks.TextBlock(label='text block')), ('test_emailblock', wagtail.core.blocks.EmailBlock()), ('test_urlblock', wagtail.core.blocks.URLBlock()), ('test_richtextblock', wagtail.core.blocks.RichTextBlock()), ('test_rawhtmlblock', wagtail.core.blocks.RawHTMLBlock()), ('test_blockquoteblock', wagtail.core.blocks.BlockQuoteBlock()), ('test_structblock', wagtail.core.blocks.StructBlock([('field_a', wagtail.core.blocks.TextBlock()), ('field_b', wagtail.core.blocks.TextBlock())])), ('test_listblock', wagtail.core.blocks.ListBlock(wagtail.core.blocks.TextBlock())), ('test_nestedstreamblock', wagtail.core.blocks.StreamBlock([('block_a', wagtail.core.blocks.TextBlock()), ('block_b', wagtail.core.blocks.TextBlock())])), ('test_customstructblock', wagtail.core.blocks.StructBlock([('field_a', wagtail.core.blocks.TextBlock()), ('field_b', wagtail.core.blocks.TextBlock())])), ('test_customblockwithoutextractmethod', wagtail_localize.test.models.CustomBlockWithoutExtractMethod()), ('test_pagechooserblock', wagtail.core.blocks.PageChooserBlock()), ('test_pagechooserblock_with_restricted_types', wagtail.core.blocks.PageChooserBlock(['wagtail_localize_test.TestHomePage', 'wagtail_localize_test.TestPage'])), ('test_imagechooserblock', wagtail.images.blocks.ImageChooserBlock()), ('test_documentchooserblock', wagtail.documents.blocks.DocumentChooserBlock()), ('test_snippetchooserblock', wagtail.snippets.blocks.SnippetChooserBlock(wagtail_localize.test.models.TestSnippet)), ('test_embedblock', wagtail.embeds.blocks.EmbedBlock())], blank=True),
        ),
        migrations.AlterField(
            model_name='testpage',
            name='test_synchronized_streamfield',
            field=wagtail.core.fields.StreamField([('test_charblock', wagtail.core.blocks.CharBlock(max_length=255)), ('test_textblock', wagtail.core.blocks.TextBlock(label='text block')), ('test_emailblock', wagtail.core.blocks.EmailBlock()), ('test_urlblock', wagtail.core.blocks.URLBlock()), ('test_richtextblock', wagtail.core.blocks.RichTextBlock()), ('test_rawhtmlblock', wagtail.core.blocks.RawHTMLBlock()), ('test_blockquoteblock', wagtail.core.blocks.BlockQuoteBlock()), ('test_structblock', wagtail.core.blocks.StructBlock([('field_a', wagtail.core.blocks.TextBlock()), ('field_b', wagtail.core.blocks.TextBlock())])), ('test_listblock', wagtail.core.blocks.ListBlock(wagtail.core.blocks.TextBlock())), ('test_nestedstreamblock', wagtail.core.blocks.StreamBlock([('block_a', wagtail.core.blocks.TextBlock()), ('block_b', wagtail.core.blocks.TextBlock())])), ('test_customstructblock', wagtail.core.blocks.StructBlock([('field_a', wagtail.core.blocks.TextBlock()), ('field_b', wagtail.core.blocks.TextBlock())])), ('test_customblockwithoutextractmethod', wagtail_localize.test.models.CustomBlockWithoutExtractMethod()), ('test_pagechooserblock', wagtail.core.blocks.PageChooserBlock()), ('test_pagechooserblock_with_restricted_types', wagtail.core.blocks.PageChooserBlock(['wagtail_localize_test.TestHomePage', 'wagtail_localize_test.TestPage'])), ('test_imagechooserblock', wagtail.images.blocks.ImageChooserBlock()), ('test_documentchooserblock', wagtail.documents.blocks.DocumentChooserBlock()), ('test_snippetchooserblock', wagtail.snippets.blocks.SnippetChooserBlock(wagtail_localize.test.models.TestSnippet)), ('test_embedblock', wagtail.embeds.blocks.EmbedBlock())], blank=True),
        ),
    ]



# Generated by Django 3.0.6 on 2020-07-17 15:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0057_page_locale_fields_notnull'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='String',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data_hash', models.UUIDField()),
                ('data', models.TextField()),
                ('locale', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='source_strings', to='wagtailcore.Locale')),
            ],
            options={
                'unique_together': {('locale', 'data_hash')},
            },
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(unique=True)),
                ('template', models.TextField()),
                ('template_format', models.CharField(max_length=100)),
                ('string_count', models.PositiveIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='TranslatableObject',
            fields=[
                ('translation_key', models.UUIDField(primary_key=True, serialize=False)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
            options={
                'unique_together': {('content_type', 'translation_key')},
            },
        ),
        migrations.CreateModel(
            name='TranslationSource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_repr', models.TextField(max_length=200)),
                ('content_json', models.TextField()),
                ('created_at', models.DateTimeField()),
                ('locale', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.Locale')),
                ('object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sources', to='wagtail_localize.TranslatableObject')),
                ('specific_content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='TranslationLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('locale', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='translation_logs', to='wagtailcore.Locale')),
                ('page_revision', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.PageRevision')),
                ('source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='translation_logs', to='wagtail_localize.TranslationSource')),
            ],
        ),
        migrations.CreateModel(
            name='TranslationContext',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path_id', models.UUIDField()),
                ('path', models.TextField()),
                ('object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtail_localize.TranslatableObject')),
            ],
            options={
                'unique_together': {('object', 'path_id')},
            },
        ),
        migrations.CreateModel(
            name='TemplateSegment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField()),
                ('context', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='wagtail_localize.TranslationContext')),
                ('source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtail_localize.TranslationSource')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='segments', to='wagtail_localize.Template')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StringSegment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField()),
                ('attrs', models.TextField(blank=True)),
                ('context', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='wagtail_localize.TranslationContext')),
                ('source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtail_localize.TranslationSource')),
                ('string', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='segments', to='wagtail_localize.String')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RelatedObjectSegment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField()),
                ('context', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='wagtail_localize.TranslationContext')),
                ('object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='references', to='wagtail_localize.TranslatableObject')),
                ('source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtail_localize.TranslationSource')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StringTranslation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data', models.TextField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('context', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='translations', to='wagtail_localize.TranslationContext')),
                ('locale', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='string_translations', to='wagtailcore.Locale')),
                ('translation_of', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='wagtail_localize.String')),
            ],
            options={
                'unique_together': {('locale', 'translation_of', 'context')},
            },
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-12 18:34

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('wagtail_localize', '0007_stringtranslation_type_and_tool_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='stringtranslation',
            name='last_translated_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
    ]
